#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <gtk/gtk.h>

#include "cfg.h"

#include "guiutils.h"
#include "tool_bar.h"
#include "cdialog.h"
#include "progressdialog.h"

#include "edv_types.h"
#include "libendeavour2-base/edv_path.h"
#include "libendeavour2-base/edv_directory.h"
#include "libendeavour2-base/edv_vfs_obj.h"
#include "libendeavour2-base/edv_vfs_obj_stat.h"
#include "libendeavour2-base/edv_recycled_obj.h"
#include "libendeavour2-base/edv_recycle_bin_index.h"
#include "edv_purge_obj.h"
#include "edv_recover_obj.h"
#include "edv_recycle_bin_sync.h"
#include "edv_obj_info_match.h"
#include "edv_utils_gtk.h"
#include "edv_find_bar.h"
#include "edv_status_bar.h"
#include "edv_confirm.h"
#include "recbin.h"
#include "recbin_cb.h"
#include "recbin_op_cb.h"
#include "recbin_contents_list.h"
#include "recbin_desktop_icon.h"
#include "edv_help.h"
#include "edv_emit.h"
#include "edv_op.h"
#include "endeavour2.h"

#include "edv_cfg_list.h"
#include "config.h"


static gint EDVRecBinCreateDirectories(
	EDVCore *core,
	const gchar *path
);

void EDVRecBinOPCB(
	ToolBarItem *item, gint id, gpointer data
);
void EDVRecBinOPEnterCB(
	ToolBarItem *item, gint id, gpointer data
);
void EDVRecBinOPLeaveCB(
	ToolBarItem *item, gint id, gpointer data
);

void EDVRecBinOPClose(edv_recbin_struct *recbin);
void EDVRecBinOPExit(edv_recbin_struct *recbin);

void EDVRecBinOPSyncRecycleBin(edv_recbin_struct *recbin);

void EDVRecBinOPSyncDisks(edv_recbin_struct *recbin);
void EDVRecBinOPWriteProtect(edv_recbin_struct *recbin);
void EDVRecBinOPDeleteMethodRecycle(edv_recbin_struct *recbin);
void EDVRecBinOPDeleteMethodPurge(edv_recbin_struct *recbin);

void EDVRecBinOPRecover(edv_recbin_struct *recbin);
void EDVRecBinOPPurge(edv_recbin_struct *recbin);
void EDVRecBinOPPurgeAll(edv_recbin_struct *recbin);
void EDVRecBinOPRename(edv_recbin_struct *recbin);
void EDVRecBinOPChMod(edv_recbin_struct *recbin);
void EDVRecBinOPChOwn(edv_recbin_struct *recbin);
void EDVRecBinOPChTime(edv_recbin_struct *recbin);
void EDVRecBinOPProperties(edv_recbin_struct *recbin);

void EDVRecBinOPSelectAll(edv_recbin_struct *recbin);
void EDVRecBinOPUnselectAll(edv_recbin_struct *recbin);
void EDVRecBinOPInvertSelection(edv_recbin_struct *recbin);

void EDVRecBinOPRefresh(edv_recbin_struct *recbin);
void EDVRecBinOPRefreshAll(edv_recbin_struct *recbin);

void EDVRecBinContentsFilter(edv_recbin_struct *recbin);

void EDVRecBinMIMETypes(edv_recbin_struct *recbin);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (gint)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Creates the directory and all parent directories as needed and
 *	then reports them.
 *
 *	The path specifies the directory, if it does not exist then
 *	it and all its parent directories will be created as needed.
 */
static gint EDVRecBinCreateDirectories(
	EDVCore *core,
	const gchar *path
)
{
	GList *new_paths_list;
	const gint	status = edv_directory_create(
		path,
		TRUE,				/* Create parents */
		&new_paths_list
	),
			error_code = (int)errno;
	if(new_paths_list != NULL)
	{
		const gchar *path;
		GList *glist;
		EDVVFSObject *obj;

		/* Notify about the new directories being created */
		for(glist = new_paths_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
			path = (const gchar *)glist->data;
			if(path == NULL)
				continue;

			obj = edv_vfs_object_lstat(path);
			if(obj != NULL)
			{
				edv_emit_vfs_object_added(
					core,
					path,
					obj
				);
				edv_vfs_object_delete(obj);
			}
		}

		g_list_foreach(
			new_paths_list,
			(GFunc)g_free,
			NULL
		);
		g_list_free(new_paths_list);
	}

	errno = (int)error_code;

	return(status);
}


/*
 *	Operation id callback nexus.
 *
 *	The given client data must be a edv_recbin_opid_struct *.
 */
void EDVRecBinOPCB(
	ToolBarItem *item, gint id, gpointer data  
)
{
	GtkWidget *toplevel;
	CfgList *cfg_list;
	edv_recbin_struct *recbin;
	EDVCore *core;
	edv_recbin_opid_struct *opid = EDV_RECBIN_OPID(data);
	if(opid == NULL)
		return;

	recbin = opid->recbin;
	if((recbin == NULL) || (opid->flags & EDV_OPID_NO_OP))
		return;

	if(EDV_RECBIN_IS_PROCESSING(recbin) || (recbin->freeze_count > 0))
		return;

	toplevel = recbin->toplevel;
	core = recbin->core;
	cfg_list = core->cfg_list;

	recbin->freeze_count++;

	/* Handle by operation id code */
	switch(opid->op)
	{
	  case EDV_RECBIN_OP_NONE:
	  case EDV_RECBIN_OP_SEPARATOR:
		break;

	  case EDV_RECBIN_OP_CLOSE:
		EDVRecBinOPClose(recbin);
		break;

	  case EDV_RECBIN_OP_EXIT:
		EDVRecBinOPExit(recbin);
		break;


	  case EDV_RECBIN_OP_SYNC_DISKS:
		EDVRecBinOPSyncDisks(recbin);
		break;

	  case EDV_RECBIN_OP_HISTORY:
		EDVMapHistoryListWin(
			core,
			-1,
			core->geometry_flags,
			(core->geometry_flags != 0) ? &core->geometry : NULL,
			toplevel
		);
		break;

	  case EDV_RECBIN_OP_RUN:
		EDVMapRunDialogCommand(
			core,
			NULL,
			NULL,
			toplevel
		);
		break;

	  case EDV_RECBIN_OP_RUN_TERMINAL:
		EDVRunTerminal(
			core,
			NULL,
			NULL,
			toplevel
		);
		break;

	  case EDV_RECBIN_OP_WRITE_PROTECT:
		EDVRecBinOPWriteProtect(recbin);
		break;

	  case EDV_RECBIN_OP_DELETE_METHOD_RECYCLE:
		EDVRecBinOPDeleteMethodRecycle(recbin);
		break;

	  case EDV_RECBIN_OP_DELETE_METHOD_PURGE:
		EDVRecBinOPDeleteMethodPurge(recbin);
		break;


	  case EDV_RECBIN_OP_SYNC_RECYCLE_BIN:
		EDVRecBinOPSyncRecycleBin(recbin);
		break;


	  case EDV_RECBIN_OP_RECOVER:
		EDVRecBinOPRecover(recbin);
		break;

	  case EDV_RECBIN_OP_PURGE:
		EDVRecBinOPPurge(recbin);
		break;

	  case EDV_RECBIN_OP_PURGE_ALL:
		EDVRecBinOPPurgeAll(recbin);
		break;

	  case EDV_RECBIN_OP_RENAME:
		EDVRecBinOPRename(recbin);
		break;

	  case EDV_RECBIN_OP_CHMOD:
		EDVRecBinOPChMod(recbin);
		break;

	  case EDV_RECBIN_OP_CHOWN:
		EDVRecBinOPChOwn(recbin);
		break;

	  case EDV_RECBIN_OP_CHTIME:
		EDVRecBinOPChTime(recbin);
		break;

	  case EDV_RECBIN_OP_PROPERTIES:
		EDVRecBinOPProperties(recbin);
		break;


	  case EDV_RECBIN_OP_SELECT_ALL:
		EDVRecBinOPSelectAll(recbin);
		break;

	  case EDV_RECBIN_OP_UNSELECT_ALL:
		EDVRecBinOPUnselectAll(recbin);
		break;

	  case EDV_RECBIN_OP_INVERT_SELECTION:
		EDVRecBinOPInvertSelection(recbin);
		break;


	  case EDV_RECBIN_OP_FIND:
		EDVMapRecBinFindWin(core, recbin);
		break;


	  case EDV_RECBIN_OP_REFRESH:
		EDVRecBinOPRefresh(recbin);
		break;

	  case EDV_RECBIN_OP_REFRESH_ALL:
		EDVRecBinOPRefreshAll(recbin);
		break;

	  case EDV_RECBIN_OP_SHOW_TOOL_BAR:
		if(core != NULL)
		{
			const gboolean state = !EDV_GET_B(
				EDV_CFG_PARM_RECBIN_SHOW_TOOL_BAR
			);
			EDV_SET_B(
				EDV_CFG_PARM_RECBIN_SHOW_TOOL_BAR,
				state
			);
			edv_queue_emit_reconfigured(core);
		}
		break;

	  case EDV_RECBIN_OP_SHOW_FIND_BAR:
		if(core != NULL)
		{
			const gboolean state = !EDV_GET_B(
				EDV_CFG_PARM_RECBIN_SHOW_FIND_BAR
			);
			EDV_SET_B(
				EDV_CFG_PARM_RECBIN_SHOW_FIND_BAR,
				state
			);
			edv_queue_emit_reconfigured(core);
		}
		break;

	  case EDV_RECBIN_OP_SHOW_STATUS_BAR:
		if(core != NULL)
		{
			const gboolean state = !EDV_GET_B(
				EDV_CFG_PARM_RECBIN_SHOW_STATUS_BAR
			);
			EDV_SET_B(
				EDV_CFG_PARM_RECBIN_SHOW_STATUS_BAR,
				state
			);
			edv_queue_emit_reconfigured(core);
		}
		break;


	  case EDV_RECBIN_OP_CONTENTS_LIST_FILTER:
		EDVRecBinContentsFilter(recbin);
		break;

	  case EDV_RECBIN_OP_CONTENTS_LIST_AUTO_RESIZE_COLUMNS:
		if(core != NULL)
		{
			const gboolean state = !EDV_GET_B(
				EDV_CFG_PARM_RECBIN_CONTENTS_LIST_AUTO_RESIZE_COLUMNS
			);
			EDV_SET_B(
				EDV_CFG_PARM_RECBIN_CONTENTS_LIST_AUTO_RESIZE_COLUMNS,
				state
			);
			edv_queue_emit_reconfigured(core);
		}
		break;


	  case EDV_RECBIN_OP_MIME_TYPES:
		EDVRecBinMIMETypes(recbin);
		break;


	  case EDV_RECBIN_OP_NEW_BROWSER:
		EDVNewBrowser(
			core,
			NULL,
			NULL,
			core->geometry_flags,
			(core->geometry_flags != 0) ? &core->geometry : NULL
		);
		break;

	  case EDV_RECBIN_OP_NEW_IMBR:
		EDVNewImbr(
			core,
			NULL,
			core->geometry_flags,
			(core->geometry_flags != 0) ? &core->geometry : NULL
		);
		break;

	  case EDV_RECBIN_OP_NEW_ARCHIVER:
		EDVNewArchiver(
			core,
			NULL,
			NULL,
			core->geometry_flags,
			(core->geometry_flags != 0) ? &core->geometry : NULL
		);
		break;


	  case EDV_RECBIN_OP_OPTIONS:
		EDVMapOptionsWin(
			core,
			core->geometry_flags,
			(core->geometry_flags != 0) ? &core->geometry : NULL,
			toplevel
		);
		break;

	  case EDV_RECBIN_OP_CUSTOMIZE:
		EDVMapCustomizeWin(
			core,
			core->geometry_flags,
			(core->geometry_flags != 0) ? &core->geometry : NULL,
			toplevel
		);
		break;


	  case EDV_RECBIN_OP_HELP_CONTENTS:
		edv_help(core, "Contents", toplevel);
		break;
	  case EDV_RECBIN_OP_HELP_FILE_BROWSER:
		edv_help(core, "File Browser", toplevel);
		break;
	  case EDV_RECBIN_OP_HELP_IMAGE_BROWSER:
		edv_help(core, "Image Browser", toplevel);
		break;
	  case EDV_RECBIN_OP_HELP_ARCHIVER:
		edv_help(core, "Archiver", toplevel);
		break;
	  case EDV_RECBIN_OP_HELP_RECYCLE_BIN:
		edv_help(core, "Recycle Bin", toplevel);
		break;
	  case EDV_RECBIN_OP_HELP_KEYS_LIST:
		edv_help(core, "Keys List", toplevel);
		break;
	  case EDV_RECBIN_OP_HELP_COMMON_OPERATIONS:
		edv_help(core, "Common Operations", toplevel);
		break;
	  case EDV_RECBIN_OP_HELP_ABOUT:
		EDVMapAboutDlg(
			core,
			core->geometry_flags, 
			(core->geometry_flags != 0) ? &core->geometry : NULL,
			toplevel
		);
		break;
	}

	recbin->freeze_count--;
}

/*
 *	Operation id enter notify callback nexus.
 *
 *	The given client data must be a edv_recbin_opid_struct *.
 */
void EDVRecBinOPEnterCB(
	ToolBarItem *item, gint id, gpointer data  
)
{
	const gchar *tooltip;
	edv_recbin_struct *recbin;
	edv_recbin_opid_struct *opid = EDV_RECBIN_OPID(data);
	if(opid == NULL)
		return;

	recbin = opid->recbin;
	if(recbin == NULL)
		return;

	if(recbin->freeze_count > 0)
		return;

	tooltip = opid->tooltip;
	if(tooltip != NULL)
		edv_status_bar_message(recbin->status_bar, tooltip, FALSE);
}

/*
 *	Operation id leave notify callback nexus.
 */
void EDVRecBinOPLeaveCB(
	ToolBarItem *item, gint id, gpointer data  
)
{
	edv_recbin_struct *recbin;
	edv_recbin_opid_struct *opid = EDV_RECBIN_OPID(data);
	if(opid == NULL)
		return;

	recbin = opid->recbin;
	if(recbin == NULL)
		return;

	if(recbin->freeze_count > 0)
		return;

	edv_status_bar_message(recbin->status_bar, NULL, FALSE);
}


/*
 *	Close.
 */
void EDVRecBinOPClose(edv_recbin_struct *recbin)
{
	if(recbin == NULL)
		return;

	EDVRecBinSyncConfiguration(recbin);
	EDVRecBinUnmap(recbin);
}

/*
 *	Close all windows.
 */
void EDVRecBinOPExit(edv_recbin_struct *recbin)
{
	EDVCore *core;

	if(recbin == NULL)
		return;

	core = recbin->core;

	/* Record the Recycle Bin's current configuration */
	EDVRecBinSyncConfiguration(recbin);

	/* Unmap the Recycle Bin */
	EDVRecBinUnmap(recbin);

	/* Schedual a new pending operation on the core to close all
	 * the windows
	 */
	core->pending_flags |= EDV_CORE_PENDING_CLOSE_ALL_WINDOWS;
}


/*
 *	Sync Disks.
 */
void EDVRecBinOPSyncDisks(edv_recbin_struct *recbin)
{
	GtkWidget *sb;
	EDVCore *core;

	if(recbin == NULL)
		return;

	sb = recbin->status_bar;
	core = recbin->core;

	EDVRecBinSetBusy(recbin, TRUE);
	edv_status_bar_message(
		sb,
		"Syncing disks...",
		TRUE
	);

	/* Sync disks */
	EDVSyncEDV(core);

	edv_status_bar_message(
		sb,
		"Disk sync done",
		FALSE
	);
	edv_status_bar_progress(sb, 0.0f, FALSE);

	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *	Write Protect Toggle.
 */
void EDVRecBinOPWriteProtect(edv_recbin_struct *recbin)
{
	gboolean state;
	CfgList *cfg_list;
	EDVCore *core;

	if(recbin == NULL)
		return;

	EDVRecBinSetBusy(recbin, TRUE);

	core = recbin->core;
	cfg_list = core->cfg_list;

	/* Get the current Master Write Protect state */
	state = EDV_GET_B(EDV_CFG_PARM_WRITE_PROTECT);

	/* Toggle the Master Write Protect */
	state = !state;

	/* Set the new Master Write Protect state */
	EDV_SET_B(EDV_CFG_PARM_WRITE_PROTECT, state);

	/* Notify about the change in the Master Write Protect state */
	edv_queue_emit_master_write_protect_changed(core);

	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *	Delete Method: Recycle.
 */
void EDVRecBinOPDeleteMethodRecycle(edv_recbin_struct *recbin)
{
	CfgList *cfg_list;
	EDVCore *core;

	if(recbin == NULL)
		return;

	EDVRecBinSetBusy(recbin, TRUE);

	core = recbin->core;
	cfg_list = core->cfg_list;

	EDV_SET_I(
		EDV_CFG_PARM_DELETE_METHOD,
		EDV_DELETE_METHOD_RECYCLE
	);
	edv_queue_emit_delete_method_changed(core);

	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *	Delete Method: Purge.
 */
void EDVRecBinOPDeleteMethodPurge(edv_recbin_struct *recbin)
{
	CfgList *cfg_list;
	EDVCore *core;

	if(recbin == NULL)
		return;

	EDVRecBinSetBusy(recbin, TRUE);

	core = recbin->core;
	cfg_list = core->cfg_list;

	EDV_SET_I(
		EDV_CFG_PARM_DELETE_METHOD,
		EDV_DELETE_METHOD_PURGE
	);
	edv_queue_emit_delete_method_changed(core);

	EDVRecBinSetBusy(recbin, FALSE);
}


/*
 *	Sync Recycle Bin.
 *
 *	Compacts recycled objects and fix any errors.
 */
void EDVRecBinOPSyncRecycleBin(edv_recbin_struct *recbin)
{
	gboolean yes_to_all = FALSE;
	gint status;
	GtkWidget	*toplevel,
					*sb;
	EDVCore *core;

	if(recbin == NULL)
		return;

	if(EDV_RECBIN_IS_PROCESSING(recbin))
		return;

	toplevel = recbin->toplevel;
	core = recbin->core;
	sb = recbin->status_bar;

	/* Check and warn if write protect is enabled */
	if(EDVCheckWriteProtect(core, TRUE, toplevel))
		return;

	edv_status_bar_message(
		sb,
		"Syncing recycle bin...",
		TRUE
	);

	EDVRecBinSetBusy(recbin, TRUE);

	/* Sync the recycle bin */
	status = edv_recycle_bin_sync(
		core,
		toplevel,
		TRUE,
		TRUE,
		&yes_to_all,
		EDVRecBinStatusMessageCB, recbin,
		EDVRecBinStatusProgressCB, recbin
	);

	/* Refresh listing of recycled objects */
	EDVRecBinOPRefresh(recbin);

	EDVRecBinSetBusy(recbin, FALSE);

	edv_status_bar_message(
		sb,
		"Recycle bin sync done",
		FALSE
	);
	edv_status_bar_progress(sb, 0.0f, FALSE);
}


/*
 *	Recover callback.
 */
void EDVRecBinOPRecover(edv_recbin_struct *recbin)
{
	gboolean yes_to_all = FALSE;
	gint		status,
			nobjs,
			nobjs_recovered = 0;
	gulong index;
	gchar		*msg,
			*new_path;
	const gchar *error_msg;
	GList		*glist,
			*obj_list = NULL;
	GtkWidget *toplevel;
	GtkCList *clist;
	EDVRecycledObject *obj = NULL;
	EDVCore *core;

	if(recbin == NULL)
		return;

	toplevel = recbin->toplevel;
	clist = GTK_CLIST(recbin->contents_clist);
	core = recbin->core;

	/* Check and warn if write protect is enabled */
	if(EDVCheckWriteProtect(core, TRUE, toplevel))
		return;

#define CLEANUP_RETURN	{				\
 if(obj_list != NULL) {					\
  g_list_foreach(					\
   obj_list,						\
   (GFunc)edv_recycled_object_delete,			\
   NULL							\
  );							\
  g_list_free(obj_list);				\
 }							\
							\
 return;						\
}

	/* Get the list of selected recycled objects */
	for(glist = clist->selection;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		obj = EDV_RECYCLED_OBJECT(gtk_clist_get_row_data(
			clist,
			(gint)glist->data
		));
		if(obj == NULL)
			continue;

		obj_list = g_list_append(
			obj_list,
			edv_recycled_object_copy(obj)
		);
	}

	/* No recycled objects to recover? */
	if(obj_list == NULL)
	{
		CLEANUP_RETURN;
	}

	nobjs = g_list_length(obj_list);

	EDVRecBinSetBusy(recbin, TRUE);

	/* Confirm recover */
	status = edv_confirm_recover(
		core,
		toplevel,
		(obj != NULL) ? obj->name : NULL,
		nobjs,
		((obj != NULL) && (nobjs == 1)) ? obj->original_path : NULL
	);
	switch(status)
	{
	  case CDIALOG_RESPONSE_YES:
	  case CDIALOG_RESPONSE_YES_TO_ALL:
		break;

	  default:
		EDVRecBinSetBusy(recbin, FALSE);
		CLEANUP_RETURN;
		break;
	}

	/* Recover each recycled object to the VFS */
	status = 0;
	for(glist = obj_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		obj = EDV_RECYCLED_OBJECT(glist->data);
		if(obj == NULL)
			continue;

		index = obj->index;

		/* Check if the original location no longer exists */
		if((obj->original_path != NULL) ?
			!edv_path_exists(obj->original_path) : FALSE
		)
		{
			/* Query the user to recreate the original
			 * location
			 */
			gint	response,
				status2;
			const gchar *parent_path = obj->original_path;
			if(yes_to_all)
			{
				response = CDIALOG_RESPONSE_YES;
			}
			else
			{
				gchar *msg = g_strdup_printf(
"The original location:\n\
\n\
    %s\n\
\n\
Of:\n\
\n\
    %s\n\
\n\
No longer exists, create it?",
					parent_path,
					obj->name
				);
				edv_play_sound_question(core);
				CDialogSetTransientFor(toplevel);
				response = CDialogGetResponse(
					"Create Directory",
					msg,
					NULL,
					CDIALOG_ICON_QUESTION,
					CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_YES_TO_ALL |
						CDIALOG_BTNFLAG_CANCEL,
					CDIALOG_BTNFLAG_YES
				);
				g_free(msg);
				CDialogSetTransientFor(NULL);
			}
			switch(response)
			{
			    case CDIALOG_RESPONSE_YES_TO_ALL:
				yes_to_all = TRUE;
			    case CDIALOG_RESPONSE_YES:
				if(EDVRecBinCreateDirectories(
					core,
					parent_path
				))
				{
					const gint error_code = (gint)errno;
					gchar *msg = g_strdup_printf(
"Unable to create directory:\n\
\n\
    %s\n\
\n\
%s.",
						parent_path,
						g_strerror(error_code)
					);
					edv_play_sound_error(core);
					edv_message_error(
						"Create Directory Failed",
						msg,
						NULL,
						toplevel
					);
					g_free(msg);
					status2 = -1;
				}
				else
				{
					status2 = 0;
				}
				break;

			    default:
				status2 = -4;
				break;
			}

			/* Stop recovering if an error occured or the
			 * user aborted
			 */
			if(status2 != 0)
			{
				status = status2;
				break;
			}
		}

		/* Recover this recycled object to the VFS */
		status = edv_recover_object(
			core,
			index,			/* Recycled object to recover */
			NULL,			/* Use the original location
						 * as the recovery path */
			&new_path,
			toplevel,
			TRUE,			/* Show progress */
			TRUE,			/* Interactive */
			&yes_to_all
		);

		/* Check for errors */
		error_msg = edv_recover_object_get_error(core);
		if(!STRISEMPTY(error_msg))
		{
			/* Report the error */
			edv_play_sound_error(core);
			edv_message_error(
				"Recover Recycled Object Error",
				error_msg,
				NULL,
				toplevel
			);
		}

                /* Count the recovered object, notify about the
                 * recycled object being removed, and notify about
		 * the new recovered VFS object
		 */
		if(new_path != NULL)
		{
			EDVVFSObject *obj;
			nobjs_recovered++;
			edv_emit_recycled_object_removed(
				core,
				index
			);
			obj = edv_vfs_object_lstat(new_path);
			if(obj != NULL)
			{
				edv_emit_vfs_object_added(
					core,
					new_path,
					obj
				);
				edv_vfs_object_delete(obj);
			}
			g_free(new_path);
		}

                /* Stop recovering if the user aborted */
		if(status == -4)
			break;
	}

	/* Unmap the progress dialog which may have been mapped in
	 * the above operation
	 */
	ProgressDialogBreakQuery(TRUE);
	ProgressDialogSetTransientFor(NULL);

	/* Update the status bar message */
	if(status == -4)
		msg = g_strdup(
"Recover operation canceled"
		);
	else if(nobjs_recovered > 0)
		msg = g_strdup_printf(
			"Recovered %i %s",
			nobjs_recovered,
			(nobjs_recovered == 1) ? "recycled object" : "recycled objects" 		);
	else
		msg = g_strdup_printf(
			"Unable to recover %s",
			(nobjs == 1) ? "recycled object" : "recycled objects"
		);
	edv_status_bar_message(recbin->status_bar, msg, FALSE);
	g_free(msg);

	/* Play the "completed" sound on success */
	if(status == 0)
		edv_play_sound_completed(core);

	EDVRecBinSetBusy(recbin, FALSE);

	CLEANUP_RETURN;
#undef CLEANUP_RETURN
}

/*
 *	Purge callback.
 */
void EDVRecBinOPPurge(edv_recbin_struct *recbin)
{
	gboolean yes_to_all = FALSE;
	gint		status,
			nobjs,
			nobjs_purged = 0;
	gulong index;
	gchar *msg;
	const gchar *error_msg;
	GList		*glist,
			*obj_list = NULL;
	GtkWidget *toplevel;
	GtkCList *clist;
	EDVRecycledObject *obj = NULL;
	EDVCore *core;

	if(recbin == NULL)
		return;

	toplevel = recbin->toplevel;
	clist = GTK_CLIST(recbin->contents_clist);
	core = recbin->core;

	/* Check and warn if write protect is enabled */
	if(EDVCheckWriteProtect(core, TRUE, toplevel))
		return;

#define CLEANUP_RETURN	{				\
 if(obj_list != NULL) {					\
  g_list_foreach(					\
   obj_list,						\
   (GFunc)edv_recycled_object_delete,			\
   NULL							\
  );							\
  g_list_free(obj_list);				\
 }							\
							\
 return;						\
}

	/* Get a list of selected recycled objects indices */
	for(glist = clist->selection;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		obj = EDV_RECYCLED_OBJECT(gtk_clist_get_row_data(
			clist,
			(gint)glist->data
		));
		if(obj == NULL)
			continue;

		obj_list = g_list_append(
			obj_list,
			edv_recycled_object_copy(obj)
		);
	}

	/* No recycled objects to purge? */
	if(obj_list == NULL)
	{
		CLEANUP_RETURN;
	}

	nobjs = g_list_length(obj_list);

	EDVRecBinSetBusy(recbin, TRUE);

	/* Confirm purge */
	status = edv_confirm_purge(
		core,
		toplevel,
		(obj != NULL) ? obj->name : NULL,
		nobjs
	);
	switch(status)
	{
	    case CDIALOG_RESPONSE_YES:
	    case CDIALOG_RESPONSE_YES_TO_ALL:
		break;
	    default:
		EDVRecBinSetBusy(recbin, FALSE);
		CLEANUP_RETURN;
		break;
	}

	/* Purge the recycled objects from the Recycle Bin */
	status = 0;
	for(glist = obj_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		obj = EDV_RECYCLED_OBJECT(glist->data);
		if(obj == NULL)
			continue;

		index = obj->index;

		/* Purge this recycled object */
		status = edv_purge_object(
			core,
			index,
			toplevel,
			(nobjs > 0) ?
				((gfloat)nobjs_purged / (gfloat)nobjs) : 0.0f,
			TRUE,			/* Show progress */
			TRUE,			/* Interactive */
			&yes_to_all
		);

		/* Check for errors */
		error_msg = edv_purge_object_get_error(core);
		if(!STRISEMPTY(error_msg))
		{
			/* Report the error */
			edv_play_sound_error(core);
			edv_message_error(
				"Purge Recycled Object Error",
				error_msg,
				NULL,
				toplevel
			);
		}

		/* Count this recycled object being purged and notify
		 * about this recycled object being purged
		 */
		if(status == 0)
		{
			nobjs_purged++;
			edv_emit_recycled_object_removed(
				core,
				index
			);
		}

		/* Stop purging if the user aborted */
		if(status == -4)
			break;
	}

	/* Unmap the progress dialog which may have been mapped in
	 * the above operation
	 */
	ProgressDialogBreakQuery(TRUE);
	ProgressDialogSetTransientFor(NULL);

	/* Update the status bar message */
	if(status == -4)
		msg = g_strdup(
"Purge operation canceled"
		);
	else if(nobjs_purged > 0)
		msg = g_strdup_printf(
"Purged %i %s",
			nobjs_purged,
			(nobjs_purged == 1) ? "recycled object" : "recycled objects"
		);
	else
		msg = g_strdup_printf(
"Unable to purge %s",
			(nobjs == 1) ? "recycled object" : "recycled objects"
		);
	edv_status_bar_message(recbin->status_bar, msg, FALSE);
	g_free(msg);

	/* Play the "completed" sound on success */
	if(status == 0) 
		edv_play_sound_completed(core);


	EDVRecBinSetBusy(recbin, FALSE);

	CLEANUP_RETURN;
#undef CLEANUP_RETURN
}

/*
 *	Purge all.
 */
void EDVRecBinOPPurgeAll(edv_recbin_struct *recbin)
{
	gboolean yes_to_all = FALSE;
	gint		status,
			nobjs,
			nobjs_purged = 0;
	gulong index;
	gchar *msg;
	const gchar	*error_msg,
			*index_path;
	GList		*glist,
			*indicies_list = NULL;
	GtkWidget *toplevel;
	CfgList *cfg_list;
	EDVRecycledObject *obj;
	EDVRecycleBinIndex *rp;
	EDVCore *core;

	if(recbin == NULL)
		return;

	toplevel = recbin->toplevel;
	core = recbin->core;
	cfg_list = core->cfg_list;

	/* Check and warn if write protect is enabled */
	if(EDVCheckWriteProtect(core, TRUE, toplevel))
		return;

	/* Get the path to the recycled objects index file */
	index_path = EDV_GET_S(EDV_CFG_PARM_FILE_RECYCLE_BIN_INDEX);
	if(index_path == NULL)
		return;

#define CLEANUP_RETURN	{		\
 g_list_free(indicies_list);		\
					\
 return;				\
}

	/* Get the list of recycled objects */
	rp = edv_recycle_bin_index_open(index_path);
	for(obj = edv_recycle_bin_index_next(rp);
	    obj != NULL;
	    obj = edv_recycle_bin_index_next(rp)
	)
		indicies_list = g_list_append(
			indicies_list,
			(gpointer)obj->index
		);
	edv_recycle_bin_index_close(rp);

	/* No recycled objects to purge? */
	if(indicies_list == NULL)
	{
		CLEANUP_RETURN;
	}

	nobjs = g_list_length(indicies_list);

	EDVRecBinSetBusy(recbin, TRUE);

	/* Confirm purge */
	status = edv_confirm_purge(
		core,
		toplevel,
		NULL,
		nobjs
	);
	switch(status)
	{
	    case CDIALOG_RESPONSE_YES:
	    case CDIALOG_RESPONSE_YES_TO_ALL:
		break;
	    default:
		EDVRecBinSetBusy(recbin, FALSE);
		CLEANUP_RETURN;
		break;
	}

	/* Purge all the objects from the Recycle Bin
	 *
	 * Note that edv_purge_all_objects() is not used here because
	 * it does not allow the reporting of each recycled object
	 * purged during the purge
	 */
	status = 0;
	for(glist = indicies_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		index = (gulong)glist->data;

		/* Purge this recycled object */
		status = edv_purge_object(
			core,
			index,			/* Recycled object to purge */
			toplevel,
			(nobjs > 0) ?
				((gfloat)nobjs_purged / (gfloat)nobjs) : -1.0f,
			TRUE,			/* Show progress */
			TRUE,			/* Interactive */
			&yes_to_all
		);

		/* Check for errors */
		error_msg = edv_purge_object_get_error(core);
		if(!STRISEMPTY(error_msg))
		{
			/* Report the error */
			edv_play_sound_error(core);
			edv_message_error(
				"Purge Recycled Object Error",
				error_msg,
				NULL,
				toplevel
			);
		}

		/* Count this recycled object being purged and notify
		 * about this recycled object being purged
		 */
		if(status == 0)
		{
			nobjs_purged++;
			edv_emit_recycled_object_removed(
				core,
				index
			);
		}

		/* Stop purging if the user aborted */
		if(status == -4)
			break;
	}

	/* If no errors occured or the user did not abort then call
	 * edv_purge_all_objects() to remove any stray recycled
	 * object files and the recycled objects index file from the
	 * Recycle Bin
	 */
	if(status == 0)
	{
		/* Remove any stray recycled object files and the recycled
		 * objects index file
		 */
		status = edv_purge_all_objects(
			core,
			toplevel,
			FALSE,			/* Do not show progress */
			TRUE,			/* Interactive */
			&yes_to_all
		);

		/* Check for errors */
		error_msg = edv_purge_object_get_error(core);
		if(!STRISEMPTY(error_msg))
		{
			/* Report the error */
			edv_play_sound_error(core);
			edv_message_error(
				"Purge Recycled Object Error",
				error_msg,
				NULL,
				toplevel
			);
		}
	}

	/* Unmap the progress dialog which may have been mapped in
	 * the above operation
	 */
	ProgressDialogBreakQuery(TRUE);
	ProgressDialogSetTransientFor(NULL);

	/* Update the status bar message */
	if(status == -4)
		msg = g_strdup(
"Purge operation canceled"
		);
	else if(nobjs_purged > 0)
		msg = g_strdup_printf(
"Purged %i %s",
			nobjs_purged,
			(nobjs_purged == 1) ? "recycled object" : "recycled objects"
		);
	else
		msg = g_strdup_printf(
"Unable to purge %s",
			(nobjs == 1) ? "recycled object" : "recycled objects"
		);
	edv_status_bar_message(recbin->status_bar, msg, FALSE);
	g_free(msg);

	/* Play the "completed" sound on success */
	if(status == 0)
		edv_play_sound_completed(core);

	EDVRecBinSetBusy(recbin, FALSE);

	CLEANUP_RETURN;
#undef CLEANUP_RETURN
}

/*
 *	Rename.
 */
void EDVRecBinOPRename(edv_recbin_struct *recbin)
{
	gint row;
	GtkWidget *toplevel;
	GtkCList *clist;
	EDVCore *core;

	if(recbin == NULL)
		return;

	toplevel = recbin->toplevel;
	clist = GTK_CLIST(recbin->contents_clist);
	core = recbin->core;

	/* Check and warn if write protect is enabled */
	if(EDVCheckWriteProtect(core, TRUE, recbin->toplevel))
		return;

	/* Get the last selected row */
	row = EDVCListGetSelectedLast(clist, NULL);
	if(row > -1)
	{
		/* Map the rename prompt */
		EDVRecBinContentsPromptRename(
			recbin,
			row, -1
		);
	}
}

/*
 *	Change Permissions.
 */
void EDVRecBinOPChMod(edv_recbin_struct *recbin)
{
	GList *sel_objs_list;
	GtkWidget *toplevel;
	edv_obj_op_dlg_struct *d;
	EDVCore *core;

	if(recbin == NULL)
		return;

	EDVRecBinSetBusy(recbin, TRUE);

	toplevel = recbin->toplevel;
	core = recbin->core;

	/* Check and warn if write protect is enabled */
	if(EDVCheckWriteProtect(core, TRUE, toplevel))
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Get the object operations dialog and create it as needed */
	d = EDVGetObjOpDlg(core);
	if(d == NULL)
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Sync data to ensure that current values to operate on */
	EDVRecBinSyncData(recbin);

	/* Get the selected objects list */
	sel_objs_list = EDVRecBinGetSelectedObjectsList(recbin, FALSE);
	if(sel_objs_list == NULL)
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Map the object operations dialog to change permissions */
	EDVObjOpDlgMapValues(
		d,
		EDV_OBJ_OP_DLG_OP_CHMOD,
		EDV_LOCATION_TYPE_RECBIN,
		sel_objs_list,
		NULL,				/* No srouce directory */
		toplevel
	);

	/* Delete the selected objects list */
	g_list_free(sel_objs_list);

	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *	Change Ownership.
 */
void EDVRecBinOPChOwn(edv_recbin_struct *recbin)
{
	GList *sel_objs_list;
	GtkWidget *toplevel;
	edv_obj_op_dlg_struct *d;
	EDVCore *core;

	if(recbin == NULL)
		return;

	EDVRecBinSetBusy(recbin, TRUE);

	toplevel = recbin->toplevel;
	core = recbin->core;

	/* Check and warn if write protect is enabled */
	if(EDVCheckWriteProtect(core, TRUE, toplevel))
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Get the object operations dialog and create it as needed */
	d = EDVGetObjOpDlg(core);
	if(d == NULL)
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Sync data to ensure that current values to operate on */
	EDVRecBinSyncData(recbin);

	/* Get the selected objects list */
	sel_objs_list = EDVRecBinGetSelectedObjectsList(recbin, FALSE);
	if(sel_objs_list == NULL)
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Map the object operations dialog to change ownership */
	EDVObjOpDlgMapValues(
		d,
		EDV_OBJ_OP_DLG_OP_CHOWN,
		EDV_LOCATION_TYPE_RECBIN,
		sel_objs_list,
		NULL,				/* No srouce directory */
		toplevel
	);

	/* Delete the selected objects list */
	g_list_free(sel_objs_list);

	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *	Change Time Stamps.
 */
void EDVRecBinOPChTime(edv_recbin_struct *recbin)
{
	GList *sel_objs_list;
	GtkWidget *toplevel;
	edv_obj_op_dlg_struct *d;
	EDVCore *core;

	if(recbin == NULL)
		return;

	EDVRecBinSetBusy(recbin, TRUE);

	toplevel = recbin->toplevel;
	core = recbin->core;

	/* Check and warn if write protect is enabled */
	if(EDVCheckWriteProtect(core, TRUE, toplevel))
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Get the object operations dialog and create it as needed */
	d = EDVGetObjOpDlg(core);
	if(d == NULL)
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Sync data to ensure that current values to operate on */
	EDVRecBinSyncData(recbin);

	/* Get the selected objects list */
	sel_objs_list = EDVRecBinGetSelectedObjectsList(recbin, FALSE);
	if(sel_objs_list == NULL)
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Map the object operations dialog to change time stamps */
	EDVObjOpDlgMapValues(
		d,
		EDV_OBJ_OP_DLG_OP_CHTIME,
		EDV_LOCATION_TYPE_RECBIN,
		sel_objs_list,
		NULL,				/* No srouce directory */
		toplevel
	);

	/* Delete the selected objects list */
	g_list_free(sel_objs_list);

	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *	Properties.
 */
void EDVRecBinOPProperties(edv_recbin_struct *recbin)
{
	GtkWidget *toplevel;
	EDVRecycledObject *obj = NULL;
	EDVCore *core;

	if(recbin == NULL)
		return;

	EDVRecBinSetBusy(recbin, TRUE);

	toplevel = recbin->toplevel;
	core = recbin->core;

	EDVRecBinSyncData(recbin);

	/* Get the selected object */
	if(recbin->contents_clist_selected_row > -1)
	{
		GtkCList *clist = GTK_CLIST(recbin->contents_clist);
		GList *glist = clist->selection_end;
		if(glist != NULL)
			obj = EDV_RECYCLED_OBJECT(gtk_clist_get_row_data(
				clist,
				(gint)glist->data
			));
	}

	/* No selected object? */
	if(obj == NULL)
	{
		EDVRecBinSetBusy(recbin, FALSE);
		return;
	}

	/* Create a new Properties Dialog to display the object */
	EDVNewPropertiesDialogRecycled(
		core,
		obj->index,
		NULL,				/* Default page */
		NULL,				/* No extended properties list */
		core->geometry_flags,
		(core->geometry_flags != 0) ? &core->geometry : NULL,
		toplevel
	);

	EDVRecBinSetBusy(recbin, FALSE);
}


/*
 *	Select all.
 */
void EDVRecBinOPSelectAll(edv_recbin_struct *recbin)
{
	GtkCList *clist;
	EDVCore *core;

	if(recbin == NULL)
		return;

	clist = GTK_CLIST(recbin->contents_clist);
	core = recbin->core;

	EDVRecBinSetBusy(recbin, TRUE);

	/* Select all rows on clist */
	gtk_clist_freeze(clist);
	gtk_clist_select_all(clist);
	gtk_clist_thaw(clist);

	/* Assume highest row index as the last selected row */
	recbin->contents_clist_selected_row = clist->rows - 1;

	edv_status_bar_message(
		recbin->status_bar, "All objects selected", FALSE
	);

	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *	Unselect all.
 */
void EDVRecBinOPUnselectAll(edv_recbin_struct *recbin)
{
	GtkCList *clist;
	EDVCore *core;

	if(recbin == NULL)
		return;

	clist = GTK_CLIST(recbin->contents_clist);
	core = recbin->core;

	EDVRecBinSetBusy(recbin, TRUE);

	/* Unselect all rows on clist */
	gtk_clist_freeze(clist);
	gtk_clist_unselect_all(clist);
	gtk_clist_thaw(clist);

	/* Mark contents clist's row as unselected */
	recbin->contents_clist_selected_row = -1;

	edv_status_bar_message(
		recbin->status_bar, "All objects unselected", FALSE
	);

	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *	Invert Selection.
 */
void EDVRecBinOPInvertSelection(edv_recbin_struct *recbin)
{
	gint row, total_rows;
	GList *glist, *selection;
	GtkCList *clist;
	EDVCore *core;

	if(recbin == NULL)
		return;

	clist = GTK_CLIST(recbin->contents_clist);
	core = recbin->core;

	EDVRecBinSetBusy(recbin, TRUE);
	gtk_clist_freeze(clist);

	/* Get copy of selected rows list from clist */
	selection = (clist->selection != NULL) ?
		g_list_copy(clist->selection) : NULL;

	for(row = 0, total_rows = clist->rows;
		row < total_rows;
		row++
	)
	{
		for(glist = selection;
			glist != NULL;
			glist = g_list_next(glist)
		)
		{
			if(row == (gint)glist->data)
			{
				gtk_clist_unselect_row(clist, row, 0);
				break;
			}
		}
		/* Row not selected? */
		if(glist == NULL)
			gtk_clist_select_row(clist, row, 0);
	}

	g_list_free(selection);

	gtk_clist_thaw(clist);
	edv_status_bar_message(
		recbin->status_bar, "Selection inverted", FALSE
	);
	EDVRecBinSetBusy(recbin, FALSE);
}


/*
 *	Refresh.
 */
void EDVRecBinOPRefresh(edv_recbin_struct *recbin)
{
	GtkWidget	*toplevel,
					*sb;
	GtkCList *clist;
	CfgList *cfg_list;
	EDVCore *core;

	if(recbin == NULL)
		return;

	toplevel = recbin->toplevel;
	core = recbin->core;
	cfg_list = core->cfg_list;
	sb = recbin->status_bar;

	EDVRecBinSetBusy(recbin, TRUE);
	GUIBlockInput(toplevel, TRUE);

	/* Update the Contents GtkCList */
	clist = GTK_CLIST(recbin->contents_clist);
	if(clist != NULL)
	{
		/* Record the last scroll position */
		const gfloat	last_x = GTK_ADJUSTMENT_GET_VALUE(clist->hadjustment),
							last_y = GTK_ADJUSTMENT_GET_VALUE(clist->vadjustment);

		gtk_clist_freeze(clist);

		/* Reget the listing */
		EDVRecBinContentsGetListing(
			recbin,
			EDV_GET_B(EDV_CFG_PARM_LISTS_SHOW_PROGRESS)
		);

		gtk_clist_thaw(clist);

		/* Scroll back to the original position */
		EDVScrollCListToPosition(clist, last_x, last_y);
	}

	EDVRecBinUpdateMenus(recbin);
	edv_status_bar_message(sb, "Refreshed contents listing", FALSE);

	GUIBlockInput(toplevel, FALSE);
	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *	Refresh All.
 */
void EDVRecBinOPRefreshAll(edv_recbin_struct *recbin)
{
	const gchar *index_path;
	const CfgList *cfg_list;
	EDVCore *core;

	if(recbin == NULL)
		return;

	core = recbin->core;
	cfg_list = core->cfg_list;

	/* Get the path to the recycled objects index file */
	index_path = CFGItemListGetValueS(
		cfg_list, EDV_CFG_PARM_FILE_RECYCLE_BIN_INDEX
	);

	/* Update the number of recycled objects on the core */
	core->last_nrecycle_bin_items = edv_recycle_bin_index_get_total(
		index_path
	);

	/* Refresh the Recycle Bin Desktop Icon */
	EDVRecBinDeskIconUpdate(core->recbin_deskicon);

	/* Refresh the Recycle Bin */
	EDVRecBinOPRefresh(recbin);
}


/*
 *	Sets the contents list filter.
 */
void EDVRecBinContentsFilter(edv_recbin_struct *recbin)
{
	if(recbin == NULL)
		return;

	EDVRecBinSetBusy(recbin, TRUE);

	if(EDVQueryNameFilter(
		recbin->core,
		&recbin->contents_list_filter,
		recbin->toplevel
	))
	{                                 
		EDVRecBinSetTitle(recbin);
		EDVRecBinOPRefresh(recbin);
	}

	EDVRecBinSetBusy(recbin, FALSE);
}

/*
 *      MIME Types.
 */
void EDVRecBinMIMETypes(edv_recbin_struct *recbin)
{
	gint i;
	gchar *type_string = NULL;
	GtkWidget *toplevel;
	GtkCList *clist;
	EDVCore *core;
	if(recbin == NULL)
		return;

	toplevel = recbin->toplevel;
	clist = GTK_CLIST(recbin->contents_clist);
	core = recbin->core;

	EDVRecBinSetBusy(recbin, TRUE);

	i = EDVCListGetSelectedLast(clist, NULL);
	if(i > -1)
	{
		EDVRecycledObject *obj = EDV_RECYCLED_OBJECT(
			gtk_clist_get_row_data(clist, i)
		);
		if(obj != NULL)
		{
			(void)edv_match_object_type_string(
				core->mime_types_list,
				obj->type,
				obj->name,
				obj->permissions,
				&type_string
			);
		}
	}

	EDVMapMIMETypesListWin(
		core,
		type_string,
		core->geometry_flags,
		(core->geometry_flags != 0) ? &core->geometry : NULL,
		toplevel
	);

	g_free(type_string);

	EDVRecBinSetBusy(recbin, FALSE);
}
